@@ -259,7 +259,7 @@ premultiply_rgba (unsigned char* data,
 		  int            height,
 		  int            stride)
 {
-    unsigned char* retdata = reinterpret_cast<unsigned char*>(malloc(stride * height));
+    unsigned char* retdata = reinterpret_cast<unsigned char*>(_cairo_malloc_ab(height, stride));
     if (!retdata)
 	return NULL;
 
@@ -322,7 +322,7 @@ _cairo_beos_bitmap_to_surface (BBitmap* bitmap)
 					bitmap->BytesPerRow());
     } else {
 	premultiplied = reinterpret_cast<unsigned char*>(
-					malloc(bitmap->BytesPerRow() * height));
+					_cairo_malloc_ab(bitmap->BytesPerRow(), height));
 	if (premultiplied)
 	    memcpy(premultiplied, bits, bitmap->BytesPerRow() * height);
     }
@@ -965,7 +965,7 @@ cairo_beos_surface_create (BView* view)
  * before the surface.
  *
  * For views that draw to a bitmap (as opposed to a screen), use this function
- * rather than cairo_beos_surface_create. Not using this function WILL lead to
+ * rather than cairo_beos_surface_create(). Not using this function WILL lead to
  * incorrect behaviour.
  *
  * For now, only views that draw to the entire area of bmp are supported.
