@@ -33,6 +33,7 @@
  * Contributor(s):
  *	Carl D. Worth <cworth@cworth.org>
  *	Kristian Høgsberg <krh@redhat.com>
+ *	Chris Wilson <chris@chris-wilson.co.uk>
  */
 
 #include "cairoint.h"
@@ -84,21 +85,37 @@ convert_data_to_bytes (png_structp png, png_row_infop row_info, png_bytep data)
 }
 
 /* Use a couple of simple error callbacks that do not print anything to
- * stderr and rely on the user to check for errors via the cairo_status_t
+ * stderr and rely on the user to check for errors via the #cairo_status_t
  * return.
  */
 static void
-png_simple_error_callback (png_structp png_save_ptr,
+png_simple_error_callback (png_structp png,
 	                   png_const_charp error_msg)
 {
-    _cairo_error (CAIRO_STATUS_NO_MEMORY);
-    longjmp (png_save_ptr->jmpbuf, CAIRO_STATUS_NO_MEMORY);
+    cairo_status_t *error = png_get_error_ptr (png);
+
+    /* default to the most likely error */
+    if (*error == CAIRO_STATUS_SUCCESS)
+	*error = _cairo_error (CAIRO_STATUS_NO_MEMORY);
+
+#ifdef PNG_SETJMP_SUPPORTED
+    longjmp (png_jmpbuf (png), 1);
+#endif
+
+    /* if we get here, then we have to choice but to abort ... */
 }
 
 static void
-png_simple_warning_callback (png_structp png_save_ptr,
+png_simple_warning_callback (png_structp png,
 	                     png_const_charp error_msg)
 {
+    cairo_status_t *error = png_get_error_ptr (png);
+
+    /* default to the most likely error */
+    if (*error == CAIRO_STATUS_SUCCESS)
+	*error = _cairo_error (CAIRO_STATUS_NO_MEMORY);
+
+    /* png does not expect to abort and will try to tidy up after a warning */
 }
 
 
@@ -108,13 +125,13 @@ write_png (cairo_surface_t	*surface,
 	   void			*closure)
 {
     int i;
-    volatile cairo_status_t status = CAIRO_STATUS_SUCCESS;
+    cairo_status_t status;
     cairo_image_surface_t *image;
     void *image_extra;
     png_struct *png;
     png_info *info;
     png_time pt;
-    png_byte **rows;
+    png_byte **volatile rows = NULL;
     png_color_16 white;
     int png_color_type;
     int depth;
@@ -123,37 +140,44 @@ write_png (cairo_surface_t	*surface,
 						  &image,
 						  &image_extra);
 
-    if (status == CAIRO_STATUS_NO_MEMORY)
-        return CAIRO_STATUS_NO_MEMORY;
-    else if (status != CAIRO_STATUS_SUCCESS)
-	return CAIRO_STATUS_SURFACE_TYPE_MISMATCH;
+    if (status == CAIRO_INT_STATUS_UNSUPPORTED)
+	return _cairo_error (CAIRO_STATUS_SURFACE_TYPE_MISMATCH);
+    else if (status)
+        return status;
 
-    rows = _cairo_malloc_ab (image->height, sizeof(png_byte*));
+    /* PNG complains about "Image width or height is zero in IHDR" */
+    if (image->width == 0 || image->height == 0) {
+	status = _cairo_error (CAIRO_STATUS_WRITE_ERROR);
+	goto BAIL1;
+    }
+
+    rows = _cairo_malloc_ab (image->height, sizeof (png_byte*));
     if (rows == NULL) {
-        status = CAIRO_STATUS_NO_MEMORY;
+	status = _cairo_error (CAIRO_STATUS_NO_MEMORY);
 	goto BAIL1;
     }
 
     for (i = 0; i < image->height; i++)
 	rows[i] = (png_byte *) image->data + i * image->stride;
 
-    png = png_create_write_struct (PNG_LIBPNG_VER_STRING, NULL,
+    png = png_create_write_struct (PNG_LIBPNG_VER_STRING, &status,
 	                           png_simple_error_callback,
 	                           png_simple_warning_callback);
     if (png == NULL) {
-	status = CAIRO_STATUS_NO_MEMORY;
+	status = _cairo_error (CAIRO_STATUS_NO_MEMORY);
 	goto BAIL2;
     }
 
     info = png_create_info_struct (png);
     if (info == NULL) {
-	status = CAIRO_STATUS_NO_MEMORY;
+	status = _cairo_error (CAIRO_STATUS_NO_MEMORY);
 	goto BAIL3;
     }
 
-    status = setjmp (png_jmpbuf (png));
-    if (status)
+#ifdef PNG_SETJMP_SUPPORTED
+    if (setjmp (png_jmpbuf (png)))
 	goto BAIL3;
+#endif
 
     png_set_write_fn (png, closure, write_func, NULL);
 
@@ -175,7 +199,7 @@ write_png (cairo_surface_t	*surface,
 	png_color_type = PNG_COLOR_TYPE_GRAY;
 	break;
     default:
-	status = CAIRO_STATUS_INVALID_FORMAT;
+	status = _cairo_error (CAIRO_STATUS_INVALID_FORMAT);
 	goto BAIL3;
     }
 
@@ -187,9 +211,8 @@ write_png (cairo_surface_t	*surface,
 		  PNG_COMPRESSION_TYPE_DEFAULT,
 		  PNG_FILTER_TYPE_DEFAULT);
 
-    white.red = 0xff;
-    white.blue = 0xff;
-    white.green = 0xff;
+    white.gray = (1 << depth) - 1;
+    white.red = white.blue = white.green = white.gray;
     png_set_bKGD (png, info, &white);
 
     png_convert_from_time_t (&pt, time (NULL));
@@ -231,8 +254,12 @@ stdio_write_func (png_structp png, png_bytep data, png_size_t size)
 	size_t ret = fwrite (data, 1, size, fp);
 	size -= ret;
 	data += ret;
-	if (size && ferror (fp))
-	    png_error(png, "Write Error");
+	if (size && ferror (fp)) {
+	    cairo_status_t *error = png_get_error_ptr (png);
+	    if (*error == CAIRO_STATUS_SUCCESS)
+		*error = _cairo_error (CAIRO_STATUS_WRITE_ERROR);
+	    png_error (png, NULL);
+	}
     }
 }
 
@@ -244,11 +271,11 @@ stdio_write_func (png_structp png, png_bytep data, png_size_t size)
  * Writes the contents of @surface to a new file @filename as a PNG
  * image.
  *
- * Return value: CAIRO_STATUS_SUCCESS if the PNG file was written
- * successfully. Otherwise, CAIRO_STATUS_NO_MEMORY if memory could not
+ * Return value: %CAIRO_STATUS_SUCCESS if the PNG file was written
+ * successfully. Otherwise, %CAIRO_STATUS_NO_MEMORY if memory could not
  * be allocated for the operation or
  * CAIRO_STATUS_SURFACE_TYPE_MISMATCH if the surface does not have
- * pixel contents, or CAIRO_STATUS_WRITE_ERROR if an I/O error occurs
+ * pixel contents, or %CAIRO_STATUS_WRITE_ERROR if an I/O error occurs
  * while attempting to write the file.
  **/
 cairo_status_t
@@ -258,20 +285,32 @@ cairo_surface_write_to_png (cairo_surface_t	*surface,
     FILE *fp;
     cairo_status_t status;
 
+    if (surface->status)
+	return surface->status;
+
+    if (surface->finished)
+	return _cairo_error (CAIRO_STATUS_SURFACE_FINISHED);
+
     fp = fopen (filename, "wb");
-    if (fp == NULL)
-	return CAIRO_STATUS_WRITE_ERROR;
+    if (fp == NULL) {
+	switch (errno) {
+	case ENOMEM:
+	    return _cairo_error (CAIRO_STATUS_NO_MEMORY);
+	default:
+	    return _cairo_error (CAIRO_STATUS_WRITE_ERROR);
+	}
+    }
 
     status = write_png (surface, stdio_write_func, fp);
 
     if (fclose (fp) && status == CAIRO_STATUS_SUCCESS)
-	status = CAIRO_STATUS_WRITE_ERROR;
+	status = _cairo_error (CAIRO_STATUS_WRITE_ERROR);
 
     return status;
 }
 
 struct png_write_closure_t {
-    cairo_write_func_t	write_func;
+    cairo_write_func_t		 write_func;
     void			*closure;
 };
 
@@ -283,8 +322,12 @@ stream_write_func (png_structp png, png_bytep data, png_size_t size)
 
     png_closure = png_get_io_ptr (png);
     status = png_closure->write_func (png_closure->closure, data, size);
-    if (status)
-	png_error(png, "Write Error");
+    if (status) {
+	cairo_status_t *error = png_get_error_ptr (png);
+	if (*error == CAIRO_STATUS_SUCCESS)
+	    *error = status;
+	png_error (png, NULL);
+    }
 }
 
 /**
@@ -295,8 +338,8 @@ stream_write_func (png_structp png, png_bytep data, png_size_t size)
  *
  * Writes the image surface to the write function.
  *
- * Return value: CAIRO_STATUS_SUCCESS if the PNG file was written
- * successfully.  Otherwise, CAIRO_STATUS_NO_MEMORY is returned if
+ * Return value: %CAIRO_STATUS_SUCCESS if the PNG file was written
+ * successfully.  Otherwise, %CAIRO_STATUS_NO_MEMORY is returned if
  * memory could not be allocated for the operation,
  * CAIRO_STATUS_SURFACE_TYPE_MISMATCH if the surface does not have
  * pixel contents.
@@ -331,21 +374,21 @@ premultiply_data (png_structp   png,
     unsigned int i;
 
     for (i = 0; i < row_info->rowbytes; i += 4) {
-	uint8_t *base = &data[i];
+	uint8_t *base  = &data[i];
 	uint8_t  alpha = base[3];
 	uint32_t p;
 
 	if (alpha == 0) {
 	    p = 0;
 	} else {
-	    uint8_t  red = base[0];
+	    uint8_t  red   = base[0];
 	    uint8_t  green = base[1];
-	    uint8_t  blue = base[2];
+	    uint8_t  blue  = base[2];
 
 	    if (alpha != 0xff) {
-		red = multiply_alpha (alpha, red);
+		red   = multiply_alpha (alpha, red);
 		green = multiply_alpha (alpha, green);
-		blue = multiply_alpha (alpha, blue);
+		blue  = multiply_alpha (alpha, blue);
 	    }
 	    p = (alpha << 24) | (red << 16) | (green << 8) | (blue << 0);
 	}
@@ -353,59 +396,90 @@ premultiply_data (png_structp   png,
     }
 }
 
+/* Converts RGBx bytes to native endian xRGB */
+static void
+convert_bytes_to_data (png_structp png, png_row_infop row_info, png_bytep data)
+{
+    unsigned int i;
+
+    for (i = 0; i < row_info->rowbytes; i += 4) {
+	uint8_t *base  = &data[i];
+	uint8_t  red   = base[0];
+	uint8_t  green = base[1];
+	uint8_t  blue  = base[2];
+	uint32_t pixel;
+
+	pixel = (0xff << 24) | (red << 16) | (green << 8) | (blue << 0);
+	memcpy (base, &pixel, sizeof (uint32_t));
+    }
+}
+
 static cairo_surface_t *
 read_png (png_rw_ptr	read_func,
 	  void		*closure)
 {
-    cairo_surface_t *surface = (cairo_surface_t*) &_cairo_surface_nil;
+    cairo_surface_t *surface;
     png_struct *png = NULL;
     png_info *info;
     png_byte *data = NULL;
     png_byte **row_pointers = NULL;
-    png_uint_32 png_width, png_height, stride;
-    int depth, color_type, interlace;
+    png_uint_32 png_width, png_height;
+    int depth, color_type, interlace, stride;
     unsigned int i;
-    unsigned int pixel_size;
+    cairo_format_t format;
+    cairo_status_t status;
 
     /* XXX: Perhaps we'll want some other error handlers? */
     png = png_create_read_struct (PNG_LIBPNG_VER_STRING,
-                                  NULL,
+                                  &status,
 	                          png_simple_error_callback,
 	                          png_simple_warning_callback);
-    if (png == NULL)
+    if (png == NULL) {
+	surface = _cairo_surface_create_in_error (_cairo_error (CAIRO_STATUS_NO_MEMORY));
 	goto BAIL;
+    }
 
     info = png_create_info_struct (png);
-    if (info == NULL)
+    if (info == NULL) {
+	surface = _cairo_surface_create_in_error (_cairo_error (CAIRO_STATUS_NO_MEMORY));
 	goto BAIL;
+    }
 
     png_set_read_fn (png, closure, read_func);
 
+    status = CAIRO_STATUS_SUCCESS;
+#ifdef PNG_SETJMP_SUPPORTED
     if (setjmp (png_jmpbuf (png))) {
-	surface = (cairo_surface_t*) &_cairo_surface_nil_read_error;
+	surface = _cairo_surface_create_in_error (status);
 	goto BAIL;
     }
+#endif
 
     png_read_info (png, info);
 
     png_get_IHDR (png, info,
                   &png_width, &png_height, &depth,
                   &color_type, &interlace, NULL, NULL);
-    stride = 4 * png_width;
+    if (status) { /* catch any early warnings */
+	surface = _cairo_surface_create_in_error (status);
+	goto BAIL;
+    }
 
     /* convert palette/gray image to rgb */
     if (color_type == PNG_COLOR_TYPE_PALETTE)
         png_set_palette_to_rgb (png);
 
     /* expand gray bit depth if needed */
-    if (color_type == PNG_COLOR_TYPE_GRAY && depth < 8)
+    if (color_type == PNG_COLOR_TYPE_GRAY) {
 #if PNG_LIBPNG_VER >= 10209
         png_set_expand_gray_1_2_4_to_8 (png);
 #else
         png_set_gray_1_2_4_to_8 (png);
 #endif
+    }
+
     /* transform transparency to alpha */
-    if (png_get_valid(png, info, PNG_INFO_tRNS))
+    if (png_get_valid (png, info, PNG_INFO_tRNS))
         png_set_tRNS_to_alpha (png);
 
     if (depth == 16)
@@ -415,37 +489,66 @@ read_png (png_rw_ptr	read_func,
         png_set_packing (png);
 
     /* convert grayscale to RGB */
-    if (color_type == PNG_COLOR_TYPE_GRAY
-        || color_type == PNG_COLOR_TYPE_GRAY_ALPHA)
-        png_set_gray_to_rgb (png);
+    if (color_type == PNG_COLOR_TYPE_GRAY ||
+	color_type == PNG_COLOR_TYPE_GRAY_ALPHA)
+    {
+	png_set_gray_to_rgb (png);
+    }
 
     if (interlace != PNG_INTERLACE_NONE)
         png_set_interlace_handling (png);
 
-    png_set_filler (png, 0xff, PNG_FILLER_AFTER);
+    switch (color_type) {
+	default:
+	case PNG_COLOR_TYPE_GRAY_ALPHA:
+	case PNG_COLOR_TYPE_RGB_ALPHA:
+	    format = CAIRO_FORMAT_ARGB32;
+	    png_set_read_user_transform_fn (png, premultiply_data);
+	    break;
 
-    png_set_read_user_transform_fn (png, premultiply_data);
+	case PNG_COLOR_TYPE_GRAY:
+	case PNG_COLOR_TYPE_PALETTE:
+	case PNG_COLOR_TYPE_RGB:
+	    format = CAIRO_FORMAT_RGB24;
+	    png_set_read_user_transform_fn (png, convert_bytes_to_data);
+	    png_set_filler (png, 0xff, PNG_FILLER_AFTER);
+	    break;
+    }
 
     png_read_update_info (png, info);
 
-    pixel_size = 4;
-    data = _cairo_malloc_abc (png_height, png_width, pixel_size);
-    if (data == NULL)
+    stride = cairo_format_stride_for_width (format, png_width);
+    if (stride < 0) {
+	surface = _cairo_surface_create_in_error (_cairo_error (CAIRO_STATUS_INVALID_STRIDE));
 	goto BAIL;
+    }
 
-    row_pointers = _cairo_malloc_ab (png_height, sizeof(char *));
-    if (row_pointers == NULL)
+    data = _cairo_malloc_ab (png_height, stride);
+    if (data == NULL) {
+	surface = _cairo_surface_create_in_error (_cairo_error (CAIRO_STATUS_NO_MEMORY));
 	goto BAIL;
+    }
+
+    row_pointers = _cairo_malloc_ab (png_height, sizeof (char *));
+    if (row_pointers == NULL) {
+	surface = _cairo_surface_create_in_error (_cairo_error (CAIRO_STATUS_NO_MEMORY));
+	goto BAIL;
+    }
 
     for (i = 0; i < png_height; i++)
-        row_pointers[i] = &data[i * png_width * pixel_size];
+        row_pointers[i] = &data[i * stride];
 
     png_read_image (png, row_pointers);
     png_read_end (png, info);
 
-    surface = cairo_image_surface_create_for_data (data,
-						   CAIRO_FORMAT_ARGB32,
-						   png_width, png_height, stride);
+    if (status) { /* catch any late warnings - probably hit an error already */
+	surface = _cairo_surface_create_in_error (status);
+	goto BAIL;
+    }
+
+    surface = cairo_image_surface_create_for_data (data, format,
+						   png_width, png_height,
+						   stride);
     if (surface->status)
 	goto BAIL;
 
@@ -460,9 +563,6 @@ read_png (png_rw_ptr	read_func,
     if (png)
 	png_destroy_read_struct (&png, &info, NULL);
 
-    if (surface->status)
-	_cairo_error (surface->status);
-
     return surface;
 }
 
@@ -476,8 +576,12 @@ stdio_read_func (png_structp png, png_bytep data, png_size_t size)
 	size_t ret = fread (data, 1, size, fp);
 	size -= ret;
 	data += ret;
-	if (size && ferror (fp))
-	    png_error(png, "Read Error");
+	if (size && (feof (fp) || ferror (fp))) {
+	    cairo_status_t *error = png_get_error_ptr (png);
+	    if (*error == CAIRO_STATUS_SUCCESS)
+		*error = _cairo_error (CAIRO_STATUS_READ_ERROR);
+	    png_error (png, NULL);
+	}
     }
 }
 
@@ -505,17 +609,19 @@ cairo_image_surface_create_from_png (const char *filename)
 
     fp = fopen (filename, "rb");
     if (fp == NULL) {
+	cairo_status_t status;
 	switch (errno) {
 	case ENOMEM:
-	    _cairo_error (CAIRO_STATUS_NO_MEMORY);
-	    return (cairo_surface_t*) &_cairo_surface_nil;
+	    status = _cairo_error (CAIRO_STATUS_NO_MEMORY);
+	    break;
 	case ENOENT:
-	    _cairo_error (CAIRO_STATUS_FILE_NOT_FOUND);
-	    return (cairo_surface_t*) &_cairo_surface_nil_file_not_found;
+	    status = _cairo_error (CAIRO_STATUS_FILE_NOT_FOUND);
+	    break;
 	default:
-	    _cairo_error (CAIRO_STATUS_READ_ERROR);
-	    return (cairo_surface_t*) &_cairo_surface_nil_read_error;
+	    status = _cairo_error (CAIRO_STATUS_READ_ERROR);
+	    break;
 	}
+	return _cairo_surface_create_in_error (status);
     }
 
     surface = read_png (stdio_read_func, fp);
@@ -526,7 +632,7 @@ cairo_image_surface_create_from_png (const char *filename)
 }
 
 struct png_read_closure_t {
-    cairo_read_func_t	read_func;
+    cairo_read_func_t		 read_func;
     void			*closure;
 };
 
@@ -538,8 +644,12 @@ stream_read_func (png_structp png, png_bytep data, png_size_t size)
 
     png_closure = png_get_io_ptr (png);
     status = png_closure->read_func (png_closure->closure, data, size);
-    if (status)
-	png_error(png, "Read Error");
+    if (status) {
+	cairo_status_t *error = png_get_error_ptr (png);
+	if (*error == CAIRO_STATUS_SUCCESS)
+	    *error = status;
+	png_error (png, NULL);
+    }
 }
 
 /**
