@@ -32,8 +32,6 @@
 #define HEIGHT 16
 #define PAD 2
 
-const char	png_filename[]	= "romedalen.png";
-
 static void
 set_solid_pattern (cairo_t *cr, int x, int y)
 {
@@ -63,7 +61,7 @@ draw_mask (cairo_t *cr, int x, int y)
     x += 0.05 * WIDTH;
     y += 0.05 * HEIGHT;
 
-    mask_surface = cairo_surface_create_similar (cairo_get_target (cr),
+    mask_surface = cairo_surface_create_similar (cairo_get_group_target (cr),
 						 CAIRO_CONTENT_ALPHA,
 						 width, height);
     cr2 = cairo_create (mask_surface);
